# -*- coding: utf-8 -*-
import requests
import datetime
import upyun
import json

from .exception import UpYunServiceException, UpYunClientException


# - wsgiref.handlers.format_date_time
def httpdate_rfc1123(dt):
    '''Return a string representation of a date according to RFC 1123
    (HTTP/1.1).

    The supplied date must be in UTC.

    '''
    weekday = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'][dt.weekday()]
    month = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
             'Oct', 'Nov', 'Dec'][dt.month - 1]
    return '%s, %02d %s %04d %02d:%02d:%02d GMT' % \
        (weekday, dt.day, month, dt.year, dt.hour, dt.minute, dt.second)


# - Date Format: RFC 1123
def cur_dt():
    return httpdate_rfc1123(datetime.datetime.utcnow())


class UpYunHttp(object):
    def __init__(self, timeout, debug):
        self.timeout = timeout
        self.debug = debug
        self.session = requests.Session()
        self.user_agent = None

    # - http://docs.python-requests.org/
    def do_http_pipe(self, method, host, uri,
                     value=None, headers={}, stream=False, files=None):
        request_id, msg, err, status = [None] * 4
        url = 'http://%s%s' % (host, uri)
        requests.adapters.DEFAULT_RETRIES = 5
        headers = self.__set_headers(headers)

        if self.debug:
            with open('debug.log', 'a') as f:
                f.write('\n\n## Http request params ##\n\n')
                kwargs = {'method': method, 'host': host, 'uri': uri,
                          'value': value, 'headers': headers,
                          'stream': stream, 'files': files,
                          'timeout': self.timeout,
                          }
                f.write('\n'.join(map(lambda kv: '%s: %s'
                                  % (kv[0], kv[1]), kwargs.items())))

        try:
            resp = self.session.request(method, url, data=value,
                                        headers=headers, stream=stream,
                                        timeout=self.timeout, files=files)
            resp.encoding = 'utf-8'
            try:
                request_id = resp.headers['X-Request-Id']
            except KeyError:
                request_id = 'Unknown'
            status = resp.status_code
            if status // 100 != 2:
                msg = resp.reason or "Unknown"
                err = resp.text
                headers = resp.headers.items()

            if self.debug:
                with open('debug.log', 'a') as f:
                    f.write('\n\n## Http responds ##\n\n')
                    kwargs = {'request_id': request_id, 'status': status,
                              'msg': msg, 'err': err}
                    f.write('\n'.join(map(lambda kv: '%s: %s'
                                      % (kv[0], kv[1]), kwargs.items())))

        except requests.exceptions.ConnectionError as e:
            raise UpYunClientException(e)
        except requests.exceptions.RequestException as e:
            raise UpYunClientException(e)
        except Exception as e:
            raise UpYunClientException(e)

        if msg:
            raise UpYunServiceException(request_id, status, msg, err, headers)

        return resp

    def __make_user_agent(self):
        default = 'upyun-python-sdk/%s' % upyun.__version__
        return json.dumps('%s %s' % (
            default, requests.utils.default_user_agent()))

    def __set_headers(self, headers):
        if 'Date' not in headers:
            headers['Date'] = cur_dt()
        if 'User-Agent' not in headers:
            headers['User-Agent'] = self.__make_user_agent()
        return headers



# -*- coding: utf-8 -*-

import json
import base64

from .modules.compat import urlencode, b
from .modules.exception import UpYunClientException, UpYunServiceException
from .modules.sign import (
    decode_msg,
    make_signature,
    make_content_md5
)
from .modules.httpipe import cur_dt


class AvPretreatment(object):
    HOST = 'p0.api.upyun.com'
    PRETREAT = '/pretreatment/'
    STATUS = '/status/'
    KEYS = ['service',
            'status_code',
            'path',
            'description',
            'task_id',
            'info',
            'signature']

    def __init__(self, service, operator, password,
                 auth_server, chunksize, hp):
        self.service = service
        self.operator = operator
        self.password = password
        self.auth_server = auth_server
        self.chunksize = chunksize
        self.hp = hp

    # --- public API
    def pretreat(self, tasks, source, notify_url, app_name=None):
        data = {'service': self.service, 'source': source,
                'notify_url': notify_url, 'tasks': tasks,
                'app_name': app_name, 'accept': 'json'}
        if not app_name:
            data.pop('app_name')
        return self.__requests_pretreatment(data)

    def status(self, taskids):
        data = {}
        if type(taskids) == 'str':
            taskids = taskids.split(',')
        if type(taskids) == list and len(taskids) <= 20:
            taskids = ','.join(taskids)
        else:
            raise UpYunClientException('length of taskids should less than 20')

        data['service'] = self.service
        data['task_ids'] = taskids
        content = self.__requests_status(data)
        if type(content) == dict and 'tasks' in content:
            return content['tasks']
        raise UpYunServiceException(None, 500,
                                    'Servers except respond tasks list',
                                    'Service Error')

    # --- private API
    def __requests_pretreatment(self, data):
        method = 'POST'
        tasks = data['tasks']
        assert isinstance(tasks, list)
        data['tasks'] = decode_msg(base64.b64encode(b(json.dumps(tasks))))

        uri = self.PRETREAT
        value = urlencode(data)
        dt = cur_dt()
        md5sum = make_content_md5(b(value))
        signature = make_signature(username=self.operator,
                                   password=self.password,
                                   auth_server=self.auth_server,
                                   method=method, uri=uri,
                                   date=dt, content_md5=md5sum)
        headers = {'Authorization': signature,
                   'Content-Type': 'application/x-www-form-urlencoded',
                   'Date': dt,
                   'Content-MD5': md5sum}
        resp = self.hp.do_http_pipe(method, self.HOST, uri,
                                    headers=headers, value=value)
        return self.__handle_resp(resp)

    def __requests_status(self, data):
        method = 'GET'
        dt = cur_dt()
        data = urlencode(data)
        uri = '%s?%s' % (self.STATUS, data)
        signature = make_signature(username=self.operator,
                                   password=self.password,
                                   auth_server=self.auth_server,
                                   method=method, uri=uri, date=dt)
        headers = {'Authorization': signature,
                   'Date': dt}
        resp = self.hp.do_http_pipe(method, self.HOST, uri, headers=headers)
        return self.__handle_resp(resp)

    def __handle_resp(self, resp):
        content = None
        try:
            content = resp.json()
        except Exception as e:
            raise UpYunClientException(e)
        return content

    def __set_params_by_post(self, value):
        data = {}
        for k, v in value.items():
            if k in self.KEYS:
                data[k] = ''.join(v) if isinstance(v, list) else v
        return data



# -*- coding: utf-8 -*-
import hashlib
import os

from .rest import UpYunRest
from .form import FormUpload
from .av import AvPretreatment
from .multi import UpYunMultiUploader

from .modules.httpipe import UpYunHttp
from .modules.exception import UpYunClientException
from .modules.compat import b
from .modules.sign import make_signature

ED_LIST = ('v%d.api.upyun.com' % ed for ed in range(4))
ED_AUTO, ED_TELECOM, ED_CNC, ED_CTT = ED_LIST

DEFAULT_CHUNKSIZE = 8192


class UpYun(object):
    def __init__(self, service, username=None, password=None,
                 auth_server=None, timeout=None, endpoint=None,
                 chunksize=None, debug=False, read_timeout=None,
                 encrypt_pwd=None):
        super(UpYun, self).__init__()
        self.service = service or os.getenv('UPYUN_SERVICE')
        self.username = username or os.getenv('UPYUN_USERNAME')
        password = password or os.getenv('UPYUN_PASSWORD')
        self.password = (hashlib.md5(b(password)).hexdigest()
                         if password else encrypt_pwd)
        self.auth_server = auth_server
        self.endpoint = endpoint or ED_AUTO
        self.chunksize = chunksize or DEFAULT_CHUNKSIZE
        self.timeout = timeout or 60
        if read_timeout is not None:
            self.requests_timeout = (self.timeout, read_timeout)
        else:
            self.requests_timeout = self.timeout
        self.hp = UpYunHttp(self.requests_timeout, debug)

        self.up_rest = UpYunRest(self.service, self.username, self.password,
                                 self.auth_server, self.endpoint,
                                 self.chunksize, self.hp)
        self.av = AvPretreatment(self.service, self.username, self.password,
                                 self.auth_server, self.chunksize, self.hp)
        self.up_form = FormUpload(self.service, self.username, self.password,
                                  self.auth_server, self.endpoint, self.hp)

        if debug:
            self.__init_debug_log(service=service, username=username,
                                  password=password, auth_server=auth_server,
                                  timeout=timeout, endpoint=endpoint,
                                  chunksize=chunksize, debug=debug)

    def __init_debug_log(self, **kwargs):
        with open('debug.log', 'w') as f:
            f.write('### Running in debug mode ###\n\n\n')
            f.write('## Initial params ##\n\n')
            f.write('\n'.join(
                map(lambda kv: '%s: %s' % (kv[0], kv[1]), kwargs.items())))

    def set_endpoint(self, endpoint, host=None):
        self.up_rest.endpoint = endpoint
        self.up_rest.host = host or ED_AUTO

    def verify_signature(self, signature, uri, headers):
        data = {
            'username': self.username,
            'password': self.password,
            'method': 'POST',
            'uri': uri,
            'date': headers['Date'],
            'auth_server': self.auth_server
        }
        if 'Content-MD5' in headers:
            data['content_md5'] = headers['Content-MD5']
        return signature == make_signature(**data)

    # --- public rest API
    def usage(self, key='/'):
        return self.up_rest.usage(key)

    def move(self, src, dest):
        if not src:
            raise UpYunClientException("missing source")
        if not dest:
            raise UpYunClientException("missing destination")

        return self.up_rest.move(src, dest)

    def copy(self, src, dest):
        if not src:
            raise UpYunClientException("missing source")
        if not dest:
            raise UpYunClientException("missing destination")

        return self.up_rest.copy(src, dest)

    def put(self, key, value, checksum=False, headers=None,
            handler=None, params=None, secret=None,
            need_resume=False, store=None, reporter=None, part_size=None,
            form=False, expiration=None, **kwargs):
        if form and hasattr(value, 'fileno'):
            return self.up_form.upload(key, value, expiration, **kwargs)
        return self.up_rest.put(key, value, checksum, headers, handler,
                                params, secret, need_resume,
                                store, reporter, part_size)

    def init_multi_uploader(self, key, headers=None, part_size=None,
                            file_size=None):
        uploader = UpYunMultiUploader(self.up_rest, key, headers=headers,
                                      part_size=part_size, file_size=file_size)
        return uploader

    def get(self, key, value=None, handler=None, params=None):
        return self.up_rest.get(key, value, handler, params)

    def delete(self, key):
        self.up_rest.delete(key)

    def mkdir(self, key):
        self.up_rest.mkdir(key)

    def getlist(self, key='/', limit=None, order=None,
                begin=None):
        return self.up_rest.getlist(key, limit, order, begin)

    def get_list_with_iter(self, key='/', limit=None, order=None,
                           begin=None):
        return self.up_rest.get_list_with_iter(key, limit, order, begin)

    def iterlist(self, key='/', limit=None, order=None,
                 begin=None):
        return self.up_rest.iterlist(key, limit, order, begin)

    def getinfo(self, key):
        return self.up_rest.getinfo(key)

    def purge(self, keys, domain=None):
        return self.up_rest.purge(keys, domain)

    # --- video pretreatment API
    def pretreat(self, tasks, source, notify_url=''):
        return self.av.pretreat(tasks, source, notify_url)

    def status(self, taskids):
        return self.av.status(taskids)

    # --- depress task
    def depress(self, tasks, notify_url):
        for task in tasks:
            save_as = task.get('save_as')
            sources = task.get('sources')
            for key in (sources, save_as):
                if not isinstance(key, str) or key == '':
                    raise UpYunClientException('Given not correct %s '
                                               'in task' % key)
        return self.av.pretreat(tasks, 'upyun', notify_url, 'depress')

    # --- compress task
    def compress(self, tasks, notify_url):
        for task in tasks:
            save_as = task.get('save_as')
            sources = task.get('sources')
            if not isinstance(save_as, str) or save_as == '':
                raise UpYunClientException('Given not correct save_as in task')
            if not isinstance(sources, list) or len(sources) == 0:
                raise UpYunClientException('Given not correct sources in task')
        return self.av.pretreat(tasks, 'upyun', notify_url, 'compress')

    def put_tasks(self, tasks, notify_url, app_name):
        return self.av.pretreat(tasks, '', notify_url, app_name)


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-

from .modules.sign import make_content_md5, make_signature
from .resume import FileStore, BaseStore, BaseReporter, print_reporter
from .modules.exception import UpYunServiceException, UpYunClientException
from .upyun import UpYun, ED_AUTO, ED_TELECOM, ED_CNC, ED_CTT

# Set default logging handler to avoid "No handler found" warnings.
import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

__title__ = 'upyun'
__version__ = '2.5.5'
__author__ = 'Monkey Zhang (timebug)'
__license__ = 'MIT License: http://www.opensource.org/licenses/mit-license.php'
__copyright__ = 'Copyright 2015 UPYUN'

__all__ = [
    'UpYun', 'UpYunServiceException', 'UpYunClientException',
    'ED_AUTO', 'ED_TELECOM', 'ED_CNC', 'ED_CTT', '__version__',
    'make_signature', 'make_content_md5', 'FileStore', 'BaseStore',
    'BaseReporter', 'print_reporter', 'add_stderr_logger'
]

logging.getLogger(__name__).addHandler(NullHandler())


def add_stderr_logger(level=logging.DEBUG):
    """
    Helper for quickly adding a StreamHandler to the logger. Useful for
    debugging.

    Returns the handler after adding it.
    """
    logger = logging.getLogger(__name__)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(
        '%(asctime)s %(levelname)s %(message)s'))
    logger.addHandler(handler)
    logger.setLevel(level)
    logger.debug('Added a stderr logging handler to logger: %s' % __name__)
    return handler


# ... Clean up.
del NullHandler



# -*- coding: utf-8 -*-
from .modules.exception import UpYunClientException

class UpYunMultiUploader(object):
    """断点续传
    :param rest: upyun rest 实例
    :param key: upyun 文件名
    :param part_size: 分块上传大小
    :param part_file: 文件大小
    :param headers: 传给 `initiate_upload` 的 HTTP 头部
    """

    def __init__(self, rest, key, headers=None,
                part_size=None, file_size=None):
        if part_size and part_size%(1024*1024) != 0:
            raise UpYunClientException('part size wrong')

        self.key = key
        self.rest = rest
        self.part_size = part_size
        self.file_size = file_size
        self.headers = headers or {}
        self._init()

    def _init(self):
        # if headers is None:
        headers = self.headers
        if self.part_size:
            headers["X-Upyun-Multi-Part-Size"] = str(self.part_size)
        if self.file_size:
            headers["X-Upyun-Multi-Length"] = str(self.file_size)
        headers["X-Upyun-Multi-Stage"] = "initiate"
        headers["X-Upyun-Multi-Disorder"] = "true"
        h = self.rest.do_http_request(
            key=self.key, method="PUT", headers=headers)
        res_headers = self.rest.get_meta_headers(h)
        self.uuid = res_headers['multi-uuid']

    def upload(self, part_id, data):
        headers = {
            "X-Upyun-Multi-Stage": "upload",
            "X-Upyun-Multi-Uuid": self.uuid,
            "X-Upyun-Part-Id": str(part_id),
        }
        self.rest.do_http_request(
            key=self.key, value=data, method="PUT", headers=headers)

    def complete(self, multi_md5=None):
        headers = {
            "X-Upyun-Multi-Stage": "complete",
            "X-Upyun-Multi-Uuid": self.uuid,
        }
        if multi_md5:
            headers["X-Upyun-Multi-Md5"] = multi_md5

        h = self.rest.do_http_request(
            key=self.key, method="PUT", headers=headers)
        res_headers = self.rest.get_meta_headers(h)
        return res_headers

    def cancel(self):
        headers = {
            "X-Upyun-Multi-Uuid": self.uuid,
        }
        self.rest.do_http_request(
            key=self.key, method="DELETE", headers=headers)



# -*- coding: utf-8 -*-
import os
import time

from .modules.exception import UpYunClientException
from .modules.sign import make_policy, make_signature
from .modules.httpipe import cur_dt


class FormUpload(object):

    def __init__(self, service, username, password,
                 auth_server, endpoint, hp):
        self.service = service
        self.username = username
        self.password = password
        self.auth_server = auth_server
        self.hp = hp
        self.host = endpoint
        self.uri = '/%s/' % service

    def upload(self, key, value, expiration, **kwargs):
        expiration = expiration or 1800
        expiration += int(time.time())
        dt = cur_dt()
        data = {
            'service': self.service,
            'expiration': expiration,
            'save-key': key,
            'date': dt,
        }
        data.update(kwargs)
        policy = make_policy(data)
        signature = make_signature(
            username=self.username,
            password=self.password,
            auth_server=self.auth_server,
            method='POST',
            uri=self.uri,
            date=dt,
            policy=policy
        )
        postdata = {
            'policy': policy,
            'authorization': signature,
            'file': (os.path.basename(value.name), value),
        }
        resp = self.hp.do_http_pipe('POST', self.host, self.uri,
                                    files=postdata)
        return self.__handle_resp(resp)

    def __handle_resp(self, resp):
        content = None
        try:
            content = resp.json()
        except Exception as e:
            raise UpYunClientException(e)
        return content



# -*- coding: utf-8 -*-
import collections
import hashlib
import logging
import os
import errno

from requests.packages.urllib3.fields import guess_content_type
from .modules.sign import decode_msg

from .modules.compat import b, stringify
from .modules.exception import UpYunResumeTraceException, UpYunServiceException

try:
    import simplejson as json
except (ImportError, SyntaxError):
    import json

TEMP_DIR = '.up-python-resume'
DEFAULT_CHUNKSIZE = 8192
THRESHOLD = 5 * 1024 * 1024
PART_SIZE = 1024 * 1024
log = logging.getLogger(__name__)


class BaseStore(object):
    @staticmethod
    def get_key(service, key, filename):
        return hashlib.md5(b('{0}-{1}-{2}'.format(
            service, key, filename))).hexdigest()

    def get(self, key):
        raise NotImplementedError("get method not found")

    def set(self, key, value):
        raise NotImplementedError("set method not found")

    def delete(self, key):
        raise NotImplementedError("delete method not found")


class FileStore(BaseStore):
    """
    :param directory: 存储路径, 默认为'~/.up-python-resume/'
    """
    def __init__(self, directory=None):
        self.dir = directory or os.path.join(os.path.expanduser('~'), TEMP_DIR)

        if os.path.isdir(self.dir):
            return
        else:
            os.makedirs(self.dir)

    def get(self, key):
        file_path = os.path.join(self.dir, key)
        if not os.path.exists(file_path):
            return {}
        try:
            with open(file_path, 'r') as f:
                content = stringify(json.load(f))
        except ValueError:
            os.remove(file_path)
            return {}
        else:
            if not isinstance(content, dict):
                os.remove(file_path)
                return {}
            else:
                return content

    def set(self, key, value):
        file_path = os.path.join(self.dir, key)
        with open(file_path, 'w') as f:
            json.dump(value, f)

    def delete(self, key):
        file_path = os.path.join(self.dir, key)
        try:
            os.remove(file_path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise e


class MemoryStore(BaseStore):
    def __init__(self):
        self.dict = {}

    def get(self, key):
        if key in self.dict:
            value = self.dict[key]
            if not isinstance(value, dict):
                self.delete(key)
                return {}
            else:
                return self.dict[key]
        else:
            return {}

    def set(self, key, value):
        self.dict[key] = value

    def delete(self, key):
        self.dict.pop(key, None)


memory_store = MemoryStore()


class ResumeTrace(object):
    def __init__(self, service, key, filename, file_md5, file_size,
                 store=None):
        self.store = store if store else memory_store
        self.file_md5 = file_md5
        self.file_size = file_size
        self.store_key = self.store.get_key(service, key, filename)
        self.record = UpYunRecord(self.store.get(self.store_key))
        try:
            self.check(self.record)
        except UpYunResumeTraceException:
            self.delete()

    def check(self, record):
        if not record:
            return

        if not isinstance(record, UpYunRecord):
            raise UpYunResumeTraceException(
                msg="get value error:" + str(record))

        if not isinstance(record.next_id, int):
            raise UpYunResumeTraceException(msg="next_id error")
        elif record.next_id == -1:
            raise UpYunResumeTraceException(msg="old file recode not deleted")

        for key in ("start", "end"):
            if not isinstance(record.get(key), int):
                raise UpYunResumeTraceException(msg="{} error".format(key))

        if not isinstance(record.multi_uuid, str):
            raise UpYunResumeTraceException(msg="multi_uuid error")

        if self.file_md5 != record.file_md5:
            raise UpYunResumeTraceException(msg="file md5 changed")

        if self.file_size != record.file_size:
            raise UpYunResumeTraceException(msg="file size changed")

    def get(self):
        return self.record

    def commit(self):
        if self.record and self.record.next_id != -1:
            self.store.set(self.store_key, self.record)
        else:
            self.delete()

    def delete(self):
        self.store.delete(self.store_key)
        self.record = UpYunRecord()

    def __enter__(self):
        return self.record

    def __exit__(self, exc_type, exc_val, exc_tb):
        log.debug("{0:>20}, next_id:{1:>10}, uuid:{2}".format(
            "commit record", self.record.next_id, self.record.multi_uuid))
        self.commit()


class UpYunRecord(dict):
    def __init__(self, *arg, **kw):
        super(UpYunRecord, self).__init__(*arg, **kw)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            return None

    def __setattr__(self, key, value):
        self[key] = value


class SizedFile(object):
    def __init__(self, file_object, start, end):
        self.file_object = file_object
        self.end = end
        self.start = start
        self.size = end - start
        self.offset = 0

    def reset(self):
        self.file_object.seek(self.start, os.SEEK_SET)
        self.offset = 0

    def read(self, chunk=None):
        if self.offset >= self.size:
            return b''

        if (chunk is None or chunk < 0) or (chunk + self.offset >= self.size):
            data = self.file_object.read(self.size - self.offset)
            self.offset = self.size
            return data

        self.offset += chunk
        return self.file_object.read(chunk)

    def __len__(self):
        return self.size

    def get_md5(self, chunksize=DEFAULT_CHUNKSIZE):
        md5 = hashlib.md5()
        for chunk in iter(lambda: self.read(chunksize), b''):
            md5.update(chunk)
        self.reset()
        return md5.hexdigest()


class BaseReporter(object):
    def __call__(self, uploaded_size, total_size, done):
        raise NotImplementedError('Reporter must be callable')


class PrintReporter(BaseReporter):
    def __call__(self, uploaded_size, total_size, done):
        if not done:
            print("reporter: {0:.2f}%".format(
                uploaded_size * 1.0 / total_size))
        else:
            print("reporter: Done")


print_reporter = PrintReporter()


class UpYunResume(object):
    """断点续传
    :param rest: upyun rest 实例
    :param key: upyun 文件名
    :param f: file object
    :param file_size: 文件大小
    :param headers: 传给 `initiate_upload` 的 HTTP 头部
    :param checksum: 默认 False，表示不进行 MD5 校验
    :param store: BaseStore 实例, 默认采用 memory_store
    """

    def __init__(self, rest, key, f, file_size,
                 headers=None, checksum=False, store=None,
                 reporter=None, part_size=None):
        self.key = key
        self.rest = rest
        self.f = f
        self.file_size = file_size
        self.part_size = part_size
        if not part_size:
            self.part_size = PART_SIZE
        self.file_md5 = self.make_md5() if checksum else ""
        self.trace = ResumeTrace(self.rest.service, key, f.name,
                                 self.file_md5, file_size, store)
        self.headers = headers or {}
        self.checksum = checksum
        self.init_headers(f.name)
        self.progress_reporter = reporter

    def make_md5(self, chunksize=DEFAULT_CHUNKSIZE):
        md5 = hashlib.md5()
        for chunk in iter(lambda: self.f.read(chunksize), b''):
            md5.update(chunk)
        self.f.seek(0)
        return decode_msg(md5.hexdigest())

    def init_headers(self, filename):
        if "X-Upyun-Multi-Type" not in self.headers:
            self.headers["X-Upyun-Multi-Type"] = guess_content_type(filename)
        self.headers["X-Upyun-Multi-Length"] = str(self.trace.file_size)
        self.headers["X-Upyun-Multi-Stage"] = "initiate,upload"
        self.headers["X-Upyun-Part-Id"] = "0"
        self.headers["X-Upyun-Multi-Part-Size"] = str(self.part_size)

    def set_record(self, record, headers):
        next_id = None
        headers = headers or []
        for header in headers:
            try:
                if 'x-upyun-next-part-id' == header[0].lower():
                    next_id = int(header[1])
                    log.debug(
                        "get {next_id} from headers".format(next_id=next_id))
                    break
            except (ValueError, KeyError, AttributeError):
                pass

        if next_id is None:
            record.clear()
        elif next_id == -1:
            record.start = self.file_size
            record.end = self.file_size
        else:
            record.start = next_id * self.part_size
            if record.start + self.part_size > self.file_size:
                record.end = self.file_size
            else:
                record.end = record.start + self.part_size
            record.next_id = next_id
        return next_id

    def get_request(self, record):
        headers = {}
        if record:
            log.debug("{0:>20}, next_id:{1:>10}".format(
                "load record", record.next_id))

            if record.start != self.file_size:
                headers.update({"X-Upyun-Multi-Uuid": record.multi_uuid,
                                "X-Upyun-Part-Id": str(record.next_id),
                                "X-Upyun-Multi-Stage": "upload"})
            else:
                headers.update({"X-Upyun-Multi-Uuid": record.multi_uuid})
        else:
            log.debug("init file")
            record.update({
                "next_id": 0, "file_size": self.file_size,
                "file_md5": self.file_md5, "start": 0,
                "end": self.part_size if self.part_size < self.file_size
                else self.file_size})
            headers.update(self.headers)

        if record.end == self.file_size:
            log.debug("complete file")
            stage = headers.get("X-Upyun-Multi-Stage", "")
            headers["X-Upyun-Multi-Stage"] = \
                stage + ",complete" if stage else "complete"
            if self.checksum:
                headers["X-Upyun-Multi-MD5"] = self.file_md5
        self.f.seek(record.start, os.SEEK_SET)
        value = SizedFile(self.f, record.start, record.end)
        if self.checksum:
            headers["Content-MD5"] = value.get_md5()
        log.debug("{0:>20}, part_id:{1:>10}, uuid:{2}".format(
            "upload file", record.next_id, record.multi_uuid))
        return dict(method="PUT", headers=headers, key=self.key, value=value)

    def step(self, res, record):
        if not record.multi_uuid:
            for header in res:
                if 'x-upyun-multi-uuid' == header[0].lower():
                    record.multi_uuid = header[1]
                    break

        if record.end == self.file_size:
            record.next_id = -1
            log.debug("upload done")
            return True
        else:
            record.start = record.end
            if record.start + self.part_size > self.file_size:
                record.end = self.file_size
            else:
                record.end = record.start + self.part_size
            record.next_id += 1
            return False

    def upload(self):
        while True:
            with self.trace as record:
                req = self.get_request(record)
                try:
                    res = self.rest.do_http_request(**req)
                except UpYunServiceException as e:
                    try:
                        reason = stringify(json.loads(e.err))
                    except (ValueError, TypeError, AttributeError):
                        raise e
                    else:
                        if reason['msg'] == "part id error":
                            next_id = self.set_record(record, e.headers)
                            if next_id is None:
                                raise e
                        elif reason['msg'] == "part already complete":
                            record.start = self.file_size
                            record.end = self.file_size
                        elif reason['msg'] == "file already upload":
                            record.next_id = -1
                            return e.headers
                        elif reason['msg'] in (
                                "x-upyun-multi-uuid not found",
                                "file md5 not match"):
                            log.debug(reason['msg'])
                            record.clear()
                            raise e
                        else:
                            raise e
                else:
                    done = self.step(res, record)
                    if isinstance(
                            self.progress_reporter, collections.Callable):
                        self.progress_reporter(
                            record.start or 0, self.file_size, done)
                    if done:
                        return res



# -*- coding: utf-8 -*-

import hashlib
import base64
import hmac
import json

import requests

from .compat import b, PY3, builtin_str, bytes, str
from .exception import UpYunClientException

DEFAULT_CHUNKSIZE = 8192


def make_content_md5(value, chunksize=DEFAULT_CHUNKSIZE):
    if hasattr(value, 'fileno'):
        md5 = hashlib.md5()
        for chunk in iter(lambda: value.read(chunksize), b''):
            md5.update(chunk)
        value.seek(0)
        return md5.hexdigest()
    elif isinstance(value, bytes) or (not PY3 and
                                      isinstance(value, builtin_str)):
        return hashlib.md5(value).hexdigest()
    else:
        raise UpYunClientException('object type error')


def decode_msg(msg):
    if isinstance(msg, bytes):
        msg = msg.decode('utf-8')
    return msg


def encode_msg(msg):
    if isinstance(msg, str):
        msg = msg.encode('utf-8')
    return msg


def make_policy(data):
    policy = json.dumps(data)
    return base64.b64encode(b(policy))


def make_signature(**kwargs):
    """The standard signature algorithm. avaliable kwargs:
    username: operator
    password: password md5
    method: GET, POST ...
    uri: uri
    date: fmt rfc1123
    *policy: params md5
    *content_md5: content md5
    *auth_server: the remote address of authentication server
    """
    for k in kwargs:
        if isinstance(kwargs[k], bytes):
            kwargs[k] = kwargs[k].decode()

    if kwargs.get('auth_server'):
        auth_server = kwargs.pop('auth_server')
        resp = requests.post(auth_server, json=kwargs)
        return resp.text

    signarr = [kwargs['method'], kwargs['uri'], kwargs['date']]
    if kwargs.get('policy'):
        signarr.append(kwargs['policy'])
    if kwargs.get('content_md5'):
        signarr.append(kwargs['content_md5'])
    signstr = '&'.join(signarr)
    signature = base64.b64encode(
        hmac.new(b(kwargs['password']), b(signstr),
                 digestmod=hashlib.sha1).digest()
    ).decode()
    return 'UPYUN %s:%s' % (kwargs['username'], signature)


def make_purge_signature(service, username, password, uri, date):
    signstr = '&'.join([uri, service, date, password])
    signature = hashlib.md5(b(signstr)).hexdigest()
    return 'UpYun %s:%s:%s' % (service, username, signature)
